{-# LANGUAGE TypeApplications #-}
module IBClient.Connection where

import Import

import Data.Binary
import Network.Socket
import Network.Socket.ByteString

import qualified Control.Exception as E
import qualified Data.ByteString as BS
import qualified Data.ByteString.Lazy as LBS
import qualified Data.ByteString.Char8 as BS8
import qualified Data.Text as T

forkClient :: App -> IO ()
forkClient app = runRIO app $ withRunInIO $ \run ->  withSocketsDo $ do
  let refs      = twsConnectionRefs $ appRefs app
      toSend    = twsConnectionSend refs
      toRecieve = twsConnectionRecieve refs
      cStatus   = twsConnectionStatus refs
      debugSend x = run $ logDebug (display $ T.pack $ "SENT    : " <> show x)
      debugRecv x = run $ logDebug (display $ T.pack $ "RECIEVED: " <> show x)
  connHost <- readTVarIO $ twsConnectionRefsHost refs
  connPort <- readTVarIO $ twsConnectionRefsPort refs
  atomically $ modifyTVar' cStatus (const TWSConnecting)
  -- TODO: throws IO-Exeption instead of returning empty list -> handle!
  addr:_ <- getAddrInfo (Just $ defaultHints { addrFlags = [AI_NUMERICHOST, AI_NUMERICSERV], addrProtocol = 0, addrSocketType = Stream}) (Just connHost) (Just connPort)
  run $ logDebug $ displayShow  addr
  E.bracketOnError (socket (addrFamily addr) (addrSocketType addr) (addrProtocol addr)) close $ \sock -> do
    connect sock $ addrAddress addr
    let connStr = BS8.pack "API\0\0\0\0\tv100..157"
    sendAll sock connStr
    debugSend connStr
    answer <- recvAll sock
    run $ logDebug (displayShow (decode . LBS.fromStrict <$> answer :: Maybe IBGenericMessage))
    -- if we have the answer we are connected
    let idStr = LBS.toStrict $ encode $ IB_StartAPI "2" "69" -- version 2, client-id: 69
    sendAll sock idStr
    debugSend idStr
    atomically $ modifyTVar' cStatus (const TWSConnected)
    run $ logInfo $ display ("Connected to TWS" :: Text)
    let go True = do
          -- abort connection, close everything
          return ()
        go False = do
          -- race: wait for MSG in Queue or for answer on socket
          input <- race (atomically $ readTQueue toSend)
                        (recvAll sock)
          case input of
            -- we want to disconnect
            Left IBDisconnect -> go True
            Left (Msg_IB_OUT x) -> do
              let msg = LBS.toStrict $ encode x
              debugSend msg
              sendAll sock msg
            -- we lost connection
            Right Nothing     -> do
                                    atomically $ do
                                      writeTQueue toRecieve IBServerGone
                                      modifyTVar' cStatus (const TWSDisconnected)
                                    run $ logWarn $ display $ T.pack "Lost connection to TWS, reconnecting..."
                                    forkClient app
            Right (Just x)    -> do debugRecv x
                                    parseMessage x
                                      where
                                        parseMessage "" = return ()
                                        parseMessage m = do
                                          let d = decodeOrFail @IB_IN (LBS.fromStrict m)
                                          case d of
                                            Right (rest, offset, result) -> do
                                              atomically $ writeTQueue toRecieve (Msg_IB_IN result)
                                              parseMessage (LBS.toStrict rest)
                                            Left (rest, offset, err) -> do
                                              run $ logInfo (display $ T.pack $ "Could not understand message: "<> ppShow err <> " ... skipping.\nRAW: " <> show m)
                                              if m == "\NUL" then do
                                                run $ logInfo (display $ T.pack "killing NUL")
                                                parseMessage (LBS.toStrict $ LBS.tail rest)
                                              else
                                                parseMessage (LBS.toStrict rest)
          go False
    go False


recvAll :: Socket -> IO (Maybe ByteString)
recvAll s = do
    d <- recv s 4096
    let l = BS.length d
    if
       | l == 0    -> return Nothing
       | l < 4096  -> return $ Just d
       | l == 4096 -> do
                        next <- recvAll s
                        return $ (d<>) <$> next
       | otherwise -> error "recvAll: recv got more bytes then requested. Impossible according to RFC"
