{-# LANGUAGE NoImplicitPrelude #-}
{-# LANGUAGE DeriveAnyClass #-}
{-# LANGUAGE DerivingStrategies #-}
{-# LANGUAGE DerivingVia #-}
{-# LANGUAGE TemplateHaskell #-}
{-# LANGUAGE TypeApplications #-}
{-# LANGUAGE TypeOperators #-}
{-# LANGUAGE UndecidableInstances #-}
{-# LANGUAGE QuantifiedConstraints #-}
{-# OPTIONS_GHC -Wno-orphans #-}
module Types where

import Data.Aeson hiding (Options)
import Data.Default
import Data.StateVar
import Data.Types.Injective
import Data.Time
import Data.FingerTree
import Data.Semigroup
import GHC.Generics
import SDL (Window)
import DearImGui
import RIO
import RIO.Process
import Lens.Micro.TH
import qualified Data.Text as T

import IBClient.Types

-- | Command line arguments
data Options = Options
  { optionsVerbose :: !Bool
  }

data WindowParams = WindowParams
  { _windowHeight :: Int
  , _windowWidth :: Int
  } deriving (Show, Generic, FromJSON, ToJSON)

instance Default WindowParams where
  def = WindowParams 1024 768

data TWSConnection = TWSConnection
  { _host :: Text
  , _port :: Text
  } deriving (Show, Generic, FromJSON, ToJSON)

instance Default TWSConnection where
  def = TWSConnection "127.0.0.1" "7497"

instance FromJSON LogLevel where
  parseJSON = withText "LogLevel" $ \case
    "LevelDebug"  -> return LevelDebug
    "LevelInfo"   -> return LevelInfo
    "LevelWarn"   -> return LevelWarn
    "LevelError"  -> return LevelError
    x -> fail $ T.unpack $ "encountered "<>x

instance ToJSON LogLevel where
  toJSON  LevelDebug    = String "LevelDebug"
  toJSON  LevelInfo     = String "LevelInfo"
  toJSON  LevelWarn     = String "LevelWarn"
  toJSON  LevelError    = String "LevelError"
  toJSON (LevelOther _) = String "LevelDebug"

data Settings = Settings
  { _windowParams  :: WindowParams
  , _twsConnection :: TWSConnection
  , _logLevel      :: LogLevel
  } deriving (Show, Generic, FromJSON, ToJSON)

makeLenses ''WindowParams
makeLenses ''TWSConnection
makeLenses ''Settings

instance Default Settings where
  def = Settings def def LevelWarn

data TWSConnectionStatus = TWSDisconnected
                         | TWSConnecting
                         | TWSConnected
                         deriving (Show, Eq, Enum, Bounded)

data TWSConnectionRefs = TWSConnectionRefs
  { twsConnectionRefsHost :: TVar String
  , twsConnectionRefsPort :: TVar String
  , twsConnectionStatus   :: TVar TWSConnectionStatus
  , twsConnectionSend     :: TQueue Msg_IB_OUT
  , twsConnectionRecieve  :: TQueue Msg_IB_IN
  }

instance Injective TWSConnectionStatus ImVec4 where
  to = \case
      TWSDisconnected -> ImVec4 1 0 0 1
      TWSConnecting   -> ImVec4 1 1 0 1
      TWSConnected    -> ImVec4 0 1 0 1

instance Injective TWSConnectionStatus String where
  to = \case
      TWSDisconnected -> "Not Connected"
      TWSConnecting   -> "Trying to connect..."
      TWSConnected    -> "Connected"

data DataRefs = DataRefs
              { accounts            :: TVar (HashMap Text IBAccount)
              , nextValidID         :: TVar (Maybe Int)
              , nextSymbolLookup    :: TVar Text
              , symbolLookupResults :: TVar [IBSymbolSample]
              }
mkIBAccount :: Text -> IBAccount
mkIBAccount u = IBAccount (IBAccountInfo u mempty mempty) mempty mempty

data IBAccount = IBAccount
               { _accountInfo       :: IBAccountInfo
               , _accountPortfolio  :: [IBPortfolioValue]
               , _accountStrategies :: [IBAccountStrategy]
               } deriving (Show, Eq)


data IBAccountInfo = IBAccountInfo
                   { _accountName       :: Text
                   , _accountProperties :: HashMap Text [(Text, Text)] -- (value, currency)
                   , _accountLastUpdate :: Text
                   } deriving (Show, Eq)

data IBPortfolioValue = IBPortfolioValue
                      { _contract      :: IBContract
                      , _position      :: Float
                      , _marketPrice   :: Float
                      , _marketValue   :: Float
                      , _averageCost   :: Float
                      , _unrealizedPNL :: Float
                      , _realizedPNL   :: Float
                      } deriving (Show, Eq)

newtype IBAccountStrategy = IBAccountStrategy [Int] -- Int -> contract-id for this strategy
  deriving (Show, Eq)

data IBSymbolSample = IBSymbolSample
                    { _symbolId :: Int
                    , _symbol :: Text
                    , _secType :: Text
                    , _primaryExchange :: Text
                    , _currency :: Text
                    , _derivatives :: [Text]
                    } deriving (Show, Eq)

makeLenses ''IBAccountStrategy
makeLenses ''IBAccountInfo
makeLenses ''IBAccount

data ChartSettings = ChartSettings
                   { chartResolution :: Int
                   , chartStart      :: Maybe UTCTime
                   , chartEnd        :: Maybe UTCTime
                   } deriving (Show, Eq)

defChartSettings :: ChartSettings
defChartSettings = ChartSettings 60 Nothing Nothing

-- data TimeWindow = TimeWindow
--                 { begin :: Int
--                 , end :: Int
--                 } deriving (Show, Eq)
--
-- instance Semigroup TimeWindow where
--   (TimeWindow a b) <> (TimeWindow x y) = TimeWindow (min a x) (max b y)
--
-- instance Monoid TimeWindow where
--   mempty = TimeWindow 0 86400

newtype TimePoint = TimePoint Int
  deriving Eq
  deriving newtype Show
  deriving (Semigroup, Monoid) via (Max Int)

data ChartStudies = SMA { window :: Int, value :: Float }
                  | OLHC { olhc_open :: Float, olhc_low :: Float, olhc_high :: Float, olhc_close :: Float}
                  deriving (Show, Eq)

data ChartPoint = ChartPoint
                { timeOfDay  :: TimePoint
                , pointValue :: Float
                , pointExtra :: [ChartStudies]
                } deriving (Show, Eq)

instance Measured TimePoint ChartPoint where
  measure = timeOfDay

data Chart = Chart
           { chartData       :: FingerTree TimePoint ChartPoint
           , chartHistData   :: HashMap Day (FingerTree TimePoint ChartPoint)
           , fillerThread    :: ThreadId
           , chartSettings   :: ChartSettings
           , chartCache      :: [ChartPoint]
           , lastCacheUpdate :: Maybe TimePoint
           , chartDirty      :: Bool
           } deriving (Show, Eq)

newtype InjetiveGettable a b = InjetiveGettable
                          { gettable :: TVar a
                          }

instance (Injective a b) => HasGetter (InjetiveGettable a b) b where
  get r = liftIO $ do
    (value :: a) <- get (gettable r)
    return $ Data.Types.Injective.to value

instance (Injective b a) => HasSetter (InjetiveGettable a b) b where
  t $= a = liftIO $ do
    let b = Data.Types.Injective.to a
    gettable t $= b

newtype DefaultJSON a = DefaultJSON { unDefaultJSON :: a }

instance (FromJSON a, Default a) => FromJSON (DefaultJSON a) where
    parseJSON v = DefaultJSON <$> (parseJSON v <|> pure def)

data AppRefs = AppRefs
  { twsConnectionRefs :: TWSConnectionRefs
  , currentAccount    :: TVar (Maybe Text)
  , appCharts         :: TVar (HashMap Text (TVar Chart))
  , tickerIdToSymbol  :: TVar (HashMap Int Text)
  }

data App = App
  { appSettings :: !Settings
  , appLogFunc :: !LogFunc
  , appProcessContext :: !ProcessContext
  , appOptions :: !Options
  , appWindow :: !Window
  , appRefs :: !AppRefs
  , appData :: !DataRefs
  -- Add other app-specific configuration information here
  }

instance HasLogFunc App where
  logFuncL = lens appLogFunc (\x y -> x { appLogFunc = y })
instance HasProcessContext App where
  processContextL = lens appProcessContext (\x y -> x { appProcessContext = y })
